/* -*- mode: js2; js2-basic-offset: 4; indent-tabs-mode: nil -*- */

// strict mode
"use strict";

// import modules
const ExtensionUtils = imports.misc.extensionUtils;
const Me = ExtensionUtils.getCurrentExtension();

const Mainloop = imports.mainloop;
const Lang = imports.lang;
const GLib = imports.gi.GLib;
const Gio = imports.gi.Gio;

const NAME_WEIGHT = 8;
const KEYWORD_WEIGHT = 3;
const DESCRIPTION_WEIGHT = 1;

const MIN_TOKEN_LENGTH = 2;
const MAX_TOKEN_LENGTH = 6;

const CACHE_DIR = `${GLib.get_user_cache_dir()}/${Me.metadata.uuid}`;
var indexed = Gio.File.new_for_path(`${CACHE_DIR}/indexed`).query_exists(null);

/**
 * Get AppSearchProvider from registered
 * providers. There is try/catch block
 * here because imports.ui.main is part
 * of gnome-shell and can not be used
 * in prefs.js.
 *
 * @return {Mixed}
 */
var provider = () => {
    try {
        const Main = imports.ui.main;
        const AppDisplay = imports.ui.appDisplay;

        let result = null;

        let searchController;

        if (Main.overview.viewSelector !== undefined)
            searchController = Main.overview.viewSelector;
        // GNOME 3.xx
        else
            searchController =
                Main.overview._overview.controls._searchController; // GNOME >= 40

        searchController._searchResults._providers.forEach((item) => {
            if (!result && item instanceof AppDisplay.AppSearchProvider)
                result = item;
        });

        return result;
    } catch (e) {
        return null;
    }
};

var Search = new Lang.Class({
    Name: "AppUtilsSearch",

    /**
     * Constructor
     *
     * @return {Void}
     */
    _init: function () {
        // listen object - file/monitor list
        this._listen = [...new Set(GLib.get_system_data_dirs())]
            .map((path) => Gio.File.new_for_path(path + "/applications"))
            .filter((file) => file.query_exists(null))
            .map((file) => {
                let monitor = file.monitor(Gio.FileMonitorFlags.NONE, null);

                // refresh on each directory change
                monitor.connect(
                    "changed",
                    Lang.bind(this, this._handleMonitorChanged)
                );

                return {
                    file: file,
                    monitor: monitor,
                };
            });
        this._interval = null;
        this._data = new Set();
        this._subprocCount = 0;

        this.refresh();
    },

    /**
     * Destructor
     *
     * @return {Void}
     */
    destroy: function () {
        this._listen.forEach((item) => {
            item.monitor.cancel();
        });
    },

    /**
     * Find all desktop files in path and append
     * each id/name to result object
     *
     * @param  {String} path
     * @return {Object}
     */
    _desktopFileObject: function (path) {
        // @todo - async?
        let dir = Gio.file_new_for_path(path);
        let children = dir.enumerate_children("*", 0, null);
        let result = {};

        let info;
        while ((info = children.next_file(null)) !== null) {
            let id = info.get_name();
            if (!id.match(/\.desktop$/)) continue;

            let app = Gio.DesktopAppInfo.new(id);
            if (!app || !app.should_show()) continue;

            result[app.get_id()] = {
                id: app.get_id(),
                name: app.get_name(),
                displayName: app.get_display_name(),
                description: app.get_description(),
                keywords: app.get_keywords(),
            };
        }
        children.close(null);

        return result;
    },

    /**
     * Data iterator
     *
     * @param  {Function} callback
     * @return {Void}
     */
    forEach: function (callback) {
        this._data.forEach((id) => {
            callback.call(this, id);
        });
    },

    _createIndex: function (appInfos) {
        const subproc = Gio.Subprocess.new(
            [
                "python3",
                Me.path + "/update_index.py",
                MAX_TOKEN_LENGTH.toString(),
                NAME_WEIGHT.toString(),
                KEYWORD_WEIGHT.toString(),
                DESCRIPTION_WEIGHT.toString(),
                JSON.stringify(appInfos),
            ],
            Gio.SubprocessFlags.NONE
        );
        this._subprocCount++;

        subproc.wait_async(null, (proc, result) => {
            proc.wait_finish(result);
            this._subprocCount--;

            if (this._subprocCount === 0) {
                indexed = true;

                const cacheDir = Gio.File.new_for_path(`${CACHE_DIR}/`);
                if (!cacheDir.query_exists(null)) {
                    cacheDir.make_directory_with_parents(null);
                }

                const indexedFile = Gio.File.new_for_path(
                    `${CACHE_DIR}/indexed`
                );
                if (!indexedFile.query_exists(null)) {
                    indexedFile
                        .create(Gio.FileCreateFlags.NONE, null)
                        .close(null);
                }
            }
        });
    },

    /**
     * Refresh data
     *
     * @return {Void}
     */
    refresh: function () {
        console.log("gnome-fuzzy-app-search", `updating index`);

        let desktopFile = [];
        this._listen.forEach((item) => {
            desktopFile.push(this._desktopFileObject(item.file.get_path()));
        });

        let result = {};
        desktopFile.reverse().forEach((desktopObject) => {
            for (let key in desktopObject) {
                result[key] = desktopObject[key];
                desktopObject[key];
            }
        });

        this._createIndex(Object.values(result));

        this._data = new Set(Object.keys(result));
    },

    /**
     * Get an array of substrings of str that are n long, i.e. n-grams
     *
     * @param  {String} str
     * @param  {Number} n
     * @return {Array}
     */
    _nGrams: function (str, n) {
        if (str.length < n) {
            return [];
        }

        const nGrams = [];

        for (let index = str.length - n; index >= 0; index--) {
            nGrams.unshift(str.slice(index, index + n));
        }

        return nGrams;
    },

    /**
     * Get list of application ids by string query, split into terms
     *
     * @param  {Array} query
     * @return {Array}
     */
    find: function (query) {
        const scores = {};

        const minTokenLength = Math.min(
            MIN_TOKEN_LENGTH,
            Math.min(...query.map((word) => word.length))
        );
        const maxTokenLength = Math.min(
            MAX_TOKEN_LENGTH,
            Math.max(...query.map((word) => word.length))
        );

        for (let i = minTokenLength; i <= maxTokenLength; i++) {
            query
                .map((term) =>
                    this._nGrams(term.replace(/\W+/g, "").toLowerCase(), i)
                )
                .reduce((a, b) => [...a, ...b])
                .forEach((nGram) => {
                    const nGramFile = Gio.File.new_for_path(
                        `${CACHE_DIR}/${nGram.length}/${nGram}`
                    );

                    if (nGramFile.query_exists(null)) {
                        const nGramStr = GLib.file_get_contents(
                            nGramFile.get_path()
                        )[1];
                        let nGramObj;

                        try {
                            nGramObj = JSON.parse(nGramStr);
                        } catch {
                            return;
                        }

                        for (let appId in nGramObj) {
                            scores[appId] =
                                (scores[appId] ?? 0) +
                                (i * nGramObj[appId]) /
                                    Math.sqrt(Object.keys(nGramObj).length);
                        }
                    }
                });
        }

        return (
            Object.keys(scores)
                // .filter((appId) => this._data.has(appId))
                .sort((a, b) => scores[a] - scores[b])
        );
    },

    /**
     * File monitor changed event handler
     *
     * @param  {Object} monitor
     * @param  {Object} file
     * @param  {Mixed}  otherFile
     * @param  {Number} eventType
     * @return {Void}
     */
    _handleMonitorChanged: function (monitor, file, otherFile, eventType) {
        Mainloop.source_remove(this._interval);

        // handle multiple changes as one with delay
        this._interval = Mainloop.timeout_add(
            1000,
            () => {
                this.refresh();

                this._interval = null;
                return !!this._interval;
            },
            null
        );
    },
});
