#!/usr/bin/env python3
"""
Take five arguments, a maximal token length and three indexing weights, as well as a json with some of the contents of a .desktop file.

Write index entries for the json, according to the weights and the maximal length of the n-grams (i.e. tokens) to index.
"""

from typing import Dict, List, Set, TypedDict, Any

import sys
import os
import re
import json
import hashlib

from gi.repository import Gio, GLib


CACHE_DIR: str
with open(
    os.path.join(os.path.dirname(__file__), "metadata.json"), "r"
) as metadata_file:
    metadata: Dict[str, Any] = json.load(metadata_file)
    CACHE_DIR = f"{GLib.get_user_cache_dir()}/{metadata['uuid']}"


try:
    os.mkdir(os.path.join(CACHE_DIR, "cached"))
except FileExistsError:
    pass

CACHED_HASHES: Set[str] = set(os.listdir(os.path.join(CACHE_DIR, "cached")))
OBSOLETE_HASHES: Set[str] = CACHED_HASHES.copy()


NONWORD_RE: re.Pattern = re.compile(r"\W+")


class Named(TypedDict):
    """
    Something that has a name and an ID (as a dict).

    Used as a base class for AppInfo and not much else.
    """

    id: str
    name: str


class AppInfo(Named, total=False):
    """A dict describing some contents of a .desktop file, most of them optional because .desktop file authors suck"""

    displayName: str
    keywords: str
    description: str


def hash_app_info(app_info: AppInfo) -> str:
    """Generate a unique hash for the AppInfo. Uses sha256.

    Args:
        app_info (AppInfo): Some contents of a desktop file, as an AppInfo dict

    Returns:
        str: A sha256 hash unique to this AppInfo dict (well, as unique as sha256 is, which is very)
    """
    whole_sha = hashlib.sha256()

    id_sha = hashlib.sha256()
    id_sha.update(app_info["id"].encode("utf-8"))
    whole_sha.update(id_sha.digest())

    name_sha = hashlib.sha256()
    name_sha.update(app_info["name"].encode("utf-8"))
    whole_sha.update(name_sha.digest())

    display_name_sha = hashlib.sha256()
    display_name_sha.update(app_info.get("displayName", "-").encode("utf-8"))
    whole_sha.update(display_name_sha.digest())

    desc_sha = hashlib.sha256()
    desc_sha.update(app_info.get("description", "-").encode("utf-8"))
    whole_sha.update(desc_sha.digest())

    kw_sha = hashlib.sha256()
    for keyword in app_info.get("keywords", ["-"]):
        kw_sha.update(keyword.encode("utf-8") + b"\x00")
    whole_sha.update(kw_sha.digest())

    return whole_sha.hexdigest()


def n_grams(string: str, length: int) -> Set[str]:
    """Split a string into all of its n-grams of the specified length

    Args:
        string (str): The string to split into n-grams
        length (int): The length of the desired n-grams
    """
    if len(string) < length:
        return set()

    n_gram_set: Set[str] = set()

    for index in range(len(string) - length + 1):
        n_gram_set.add(string[index : index + length])

    return n_gram_set


if any(
    help_arg in sys.argv[1:]
    for help_arg in ("-h", "-H", "--help", "help", "?", "-?")
):
    print(
        f"""Usage: {sys.argv[0]} MAX_TOKEN_LENGTH NAME_WEIGHT KEYWORD_WEIGHT DESCRIPTION_WEIGHT APP_JSON

Args:
    MAX_TOKEN_LENGTH:     The maximum length of n-grams to consider for caching
    NAME_WEIGHT:          How heavy an n-gram occurring in the app's name should be
    KEYWORDS_WEIGHT:      How heavy an n-gram occurring in the app's keywords should be
    DESCRIPTION_WEIGHT:   How heavy an n-gram occurring in the app's description should be
    APP_JSON:             A json array of objects containing the desktop files' app infos to index.
                          These must contain 'name' and 'id' and should ideally contain
                          'displayName', 'keywords' and 'description',
                          where all values are strings, except for 'keywords',
                          which is an array of strings

Since this doesn't use argparse, passing 'help', '-h', '-H', '--help', '?' and/or '-?'
at any point yields this message and exits. Like right now. You're doing it right now.
"""
    )
    sys.exit(0)

if len(sys.argv[1:]) != 5:
    print(
        f"You passed an invalid amount of arguments ({len(sys.argv[1:])} instead of {5}). Try '--help'."
    )
    sys.exit(1)


MAX_TOKEN_LENGTH: int = json.loads(sys.argv[1])

NAME_WEIGHT: int = json.loads(sys.argv[2])
KEYWORD_WEIGHT: int = json.loads(sys.argv[3])
DESCRIPTION_WEIGHT: int = json.loads(sys.argv[4])

app_infos: List[AppInfo] = json.loads(sys.argv[5])


def add_to_index(app_info: AppInfo) -> None:
    """Generate all n-grams in the app info and save occurrences to the index, weighted.

    Args:
        app_info (AppInfo): An AppInfo dict with some contents of a desktop file
    """
    for key in [*app_info.keys()]:
        if app_info[key] is None:  # type: ignore[misc]
            del app_info[key]  # type: ignore[misc]

    app_info_hash = hash_app_info(app_info)

    if app_info_hash in CACHED_HASHES:
        OBSOLETE_HASHES.remove(app_info_hash)
        return

    for i in range(1, MAX_TOKEN_LENGTH + 1):
        occurrences: Dict[str, int] = {}

        for n_gram in set().union(
            *[
                n_grams(name_word.lower(), i)
                for name_word in NONWORD_RE.split(
                    app_info.get("displayName", app_info["name"])
                )
                if len(name_word) > 0
            ],
        ):
            occurrences[n_gram] = occurrences.get(n_gram, 0) + NAME_WEIGHT * i

        for n_gram in set().union(
            *[
                n_grams(NONWORD_RE.sub(keyword, "").lower(), i)
                for keyword in app_info.get("keywords", [])
                if len(keyword) > 0
            ],
        ):
            occurrences[n_gram] = (
                occurrences.get(n_gram, 0) + KEYWORD_WEIGHT * i
            )

        for n_gram in set().union(
            *[
                n_grams(desc_word.lower(), i)
                for desc_word in NONWORD_RE.split(
                    app_info.get("description", "")
                )
                if len(desc_word) > 0
            ],
        ):
            occurrences[n_gram] = (
                occurrences.get(n_gram, 0) + DESCRIPTION_WEIGHT * i
            )

        n_cache_dir = Gio.File.new_for_path(f"{CACHE_DIR}/{i}/")
        if not n_cache_dir.query_exists(None):
            n_cache_dir.make_directory_with_parents(None)

        for n_gram, weight in occurrences.items():
            n_gram_file = Gio.File.new_for_path(f"{CACHE_DIR}/{i}/{n_gram}")

            n_gram_obj: Dict[str, int]

            if not n_gram_file.query_exists(None):
                n_gram_obj = {}
                n_gram_obj[app_info["id"]] = weight

                output_stream = n_gram_file.create(
                    Gio.FileCreateFlags.NONE, None
                )
                output_stream.write(
                    json.dumps(n_gram_obj).encode("utf-8"), None
                )
                output_stream.close(None)
            else:
                n_gram_str = GLib.file_get_contents(n_gram_file.get_path())[1]

                try:
                    n_gram_obj = json.loads(n_gram_str)
                except json.JSONDecodeError:
                    n_gram_obj = {}
                    n_gram_obj[app_info["id"]] = weight

                    output_stream = n_gram_file.create(
                        Gio.FileCreateFlags.NONE, None
                    )
                    output_stream.write(
                        json.dumps(n_gram_obj).encode("utf-8"), None
                    )
                    output_stream.close(None)

                    continue

                n_gram_obj[app_info["id"]] = weight

                output_stream = n_gram_file.replace(
                    None, False, Gio.FileCreateFlags.NONE, None
                )
                output_stream.write(
                    json.dumps(n_gram_obj).encode("utf-8"), None
                )
                output_stream.close(None)

    CACHED_HASHES.add(app_info_hash)
    with open(
        os.path.join(CACHE_DIR, "cached", app_info_hash), "w"
    ) as cached_hash_file:
        json.dump(app_info, cached_hash_file)


def remove_from_index(app_info: AppInfo) -> None:
    """Remove all mentions of the specified app from the index

    Args:
        app_info (AppInfo): An AppInfo dict with some contents of a desktop file
    """
    for key in [*app_info.keys()]:
        if app_info[key] is None:  # type: ignore[misc]
            del app_info[key]  # type: ignore[misc]

    app_info_hash = hash_app_info(app_info)

    for i in range(1, MAX_TOKEN_LENGTH + 1):
        app_n_grams: Set[str] = set()

        for n_gram in set().union(
            *[
                n_grams(name_word.lower(), i)
                for name_word in NONWORD_RE.split(
                    app_info.get("displayName", app_info["name"])
                )
                if len(name_word) > 0
            ],
        ):
            app_n_grams.add(n_gram)

        for n_gram in set().union(
            *[
                n_grams(NONWORD_RE.sub(keyword, "").lower(), i)
                for keyword in app_info.get("keywords", [])
                if len(keyword) > 0
            ],
        ):
            app_n_grams.add(n_gram)

        for n_gram in set().union(
            *[
                n_grams(desc_word.lower(), i)
                for desc_word in NONWORD_RE.split(
                    app_info.get("description", "")
                )
                if len(desc_word) > 0
            ],
        ):
            app_n_grams.add(n_gram)

        for n_gram in app_n_grams:
            n_gram_file = Gio.File.new_for_path(f"{CACHE_DIR}/{i}/{n_gram}")

            n_gram_obj: Dict[str, int]

            n_gram_str = GLib.file_get_contents(n_gram_file.get_path())[1]

            n_gram_obj = json.loads(n_gram_str)

            if app_info["id"] in n_gram_obj:
                del n_gram_obj[app_info["id"]]

            output_stream = n_gram_file.replace(
                None, False, Gio.FileCreateFlags.NONE, None
            )
            output_stream.write(json.dumps(n_gram_obj).encode("utf-8"), None)
            output_stream.close(None)

    CACHED_HASHES.remove(app_info_hash)
    os.remove(os.path.join(CACHE_DIR, "cached", app_info_hash))


for app_info in app_infos:
    add_to_index(app_info)

for obsolete_hash in OBSOLETE_HASHES:
    with open(
        os.path.join(CACHE_DIR, "cached", obsolete_hash), "r"
    ) as obsolete_hash_file:
        app_info = json.load(obsolete_hash_file)

    remove_from_index(app_info)
